# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.chat_completion_response import ChatCompletionResponse
from ..types.function import Function
from ..types.function_call import FunctionCall
from ..types.message import Message
from ..types.response_format import ResponseFormat
from ..types.stream_options import StreamOptions
from ..types.tool import Tool
from .raw_client import AsyncRawChatCompletionsClient, RawChatCompletionsClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class ChatCompletionsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawChatCompletionsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawChatCompletionsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawChatCompletionsClient
        """
        return self._raw_client

    def create_chat_completions(
        self,
        *,
        model: typing.Optional[str] = OMIT,
        messages: typing.Optional[typing.Sequence[Message]] = OMIT,
        temperature: typing.Optional[float] = OMIT,
        top_p: typing.Optional[float] = OMIT,
        n: typing.Optional[int] = OMIT,
        stream: typing.Optional[bool] = OMIT,
        stream_options: typing.Optional[StreamOptions] = OMIT,
        stop: typing.Optional[typing.Sequence[str]] = OMIT,
        max_tokens: typing.Optional[int] = OMIT,
        max_completion_tokens: typing.Optional[int] = OMIT,
        presence_penalty: typing.Optional[float] = OMIT,
        frequency_penalty: typing.Optional[float] = OMIT,
        logit_bias: typing.Optional[typing.Dict[str, int]] = OMIT,
        user: typing.Optional[str] = OMIT,
        response_format: typing.Optional[ResponseFormat] = OMIT,
        seed: typing.Optional[int] = OMIT,
        tools: typing.Optional[typing.Sequence[Tool]] = OMIT,
        tool_choice: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        parallel_tool_calls: typing.Optional[bool] = OMIT,
        functions: typing.Optional[typing.Sequence[Function]] = OMIT,
        function_call: typing.Optional[FunctionCall] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ChatCompletionResponse:
        """
        Create chat completions

        Parameters
        ----------
        model : typing.Optional[str]

        messages : typing.Optional[typing.Sequence[Message]]

        temperature : typing.Optional[float]

        top_p : typing.Optional[float]

        n : typing.Optional[int]

        stream : typing.Optional[bool]

        stream_options : typing.Optional[StreamOptions]

        stop : typing.Optional[typing.Sequence[str]]

        max_tokens : typing.Optional[int]

        max_completion_tokens : typing.Optional[int]

        presence_penalty : typing.Optional[float]

        frequency_penalty : typing.Optional[float]

        logit_bias : typing.Optional[typing.Dict[str, int]]

        user : typing.Optional[str]

        response_format : typing.Optional[ResponseFormat]

        seed : typing.Optional[int]

        tools : typing.Optional[typing.Sequence[Tool]]

        tool_choice : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]

        parallel_tool_calls : typing.Optional[bool]

        functions : typing.Optional[typing.Sequence[Function]]

        function_call : typing.Optional[FunctionCall]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ChatCompletionResponse
            Chat completions response

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.chat_completions.create_chat_completions()
        """
        _response = self._raw_client.create_chat_completions(
            model=model,
            messages=messages,
            temperature=temperature,
            top_p=top_p,
            n=n,
            stream=stream,
            stream_options=stream_options,
            stop=stop,
            max_tokens=max_tokens,
            max_completion_tokens=max_completion_tokens,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            user=user,
            response_format=response_format,
            seed=seed,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            functions=functions,
            function_call=function_call,
            request_options=request_options,
        )
        return _response.data


class AsyncChatCompletionsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawChatCompletionsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawChatCompletionsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawChatCompletionsClient
        """
        return self._raw_client

    async def create_chat_completions(
        self,
        *,
        model: typing.Optional[str] = OMIT,
        messages: typing.Optional[typing.Sequence[Message]] = OMIT,
        temperature: typing.Optional[float] = OMIT,
        top_p: typing.Optional[float] = OMIT,
        n: typing.Optional[int] = OMIT,
        stream: typing.Optional[bool] = OMIT,
        stream_options: typing.Optional[StreamOptions] = OMIT,
        stop: typing.Optional[typing.Sequence[str]] = OMIT,
        max_tokens: typing.Optional[int] = OMIT,
        max_completion_tokens: typing.Optional[int] = OMIT,
        presence_penalty: typing.Optional[float] = OMIT,
        frequency_penalty: typing.Optional[float] = OMIT,
        logit_bias: typing.Optional[typing.Dict[str, int]] = OMIT,
        user: typing.Optional[str] = OMIT,
        response_format: typing.Optional[ResponseFormat] = OMIT,
        seed: typing.Optional[int] = OMIT,
        tools: typing.Optional[typing.Sequence[Tool]] = OMIT,
        tool_choice: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = OMIT,
        parallel_tool_calls: typing.Optional[bool] = OMIT,
        functions: typing.Optional[typing.Sequence[Function]] = OMIT,
        function_call: typing.Optional[FunctionCall] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> ChatCompletionResponse:
        """
        Create chat completions

        Parameters
        ----------
        model : typing.Optional[str]

        messages : typing.Optional[typing.Sequence[Message]]

        temperature : typing.Optional[float]

        top_p : typing.Optional[float]

        n : typing.Optional[int]

        stream : typing.Optional[bool]

        stream_options : typing.Optional[StreamOptions]

        stop : typing.Optional[typing.Sequence[str]]

        max_tokens : typing.Optional[int]

        max_completion_tokens : typing.Optional[int]

        presence_penalty : typing.Optional[float]

        frequency_penalty : typing.Optional[float]

        logit_bias : typing.Optional[typing.Dict[str, int]]

        user : typing.Optional[str]

        response_format : typing.Optional[ResponseFormat]

        seed : typing.Optional[int]

        tools : typing.Optional[typing.Sequence[Tool]]

        tool_choice : typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]]

        parallel_tool_calls : typing.Optional[bool]

        functions : typing.Optional[typing.Sequence[Function]]

        function_call : typing.Optional[FunctionCall]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        ChatCompletionResponse
            Chat completions response

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.chat_completions.create_chat_completions()
        asyncio.run(main())
        """
        _response = await self._raw_client.create_chat_completions(
            model=model,
            messages=messages,
            temperature=temperature,
            top_p=top_p,
            n=n,
            stream=stream,
            stream_options=stream_options,
            stop=stop,
            max_tokens=max_tokens,
            max_completion_tokens=max_completion_tokens,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
            logit_bias=logit_bias,
            user=user,
            response_format=response_format,
            seed=seed,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            functions=functions,
            function_call=function_call,
            request_options=request_options,
        )
        return _response.data



import logging
import json
import math
from typing import TYPE_CHECKING
from opik import exceptions
from opik.evaluation.metrics import score_result
from opik.evaluation.metrics.llm_judges import parsing_helpers
from opik.logging_messages import GEVAL_SCORE_CALC_FAILED

if TYPE_CHECKING:  # TODO: Daniel check if this is needed
    from litellm.types.utils import ModelResponse

LOGGER = logging.getLogger(__name__)


def parse_model_output(
    content: "ModelResponse", name: str, log_probs_supported: bool
) -> score_result.ScoreResult:
    """
    This method computes the final score based on the model's response. The model's response is a dictionary
    with a `score` key and a `reason` key. The prompt template also specifies that the score should be an integer
    between 0 and 10.

    In order to make the score computation more robust, we look at the top logprobs of the score token and compute
    a weighted average of the scores. Since we try to enforce the format of the model's response, we can assume that
    the score token is always the fourth token in the response (first token is `{"`, followed by `score` and `":`).
    """
    try:
        if not log_probs_supported:
            dict_content = parsing_helpers.extract_json_content_or_raise(
                content.choices[0].message.content
            )

            score = float(dict_content["score"])
            if not 0 <= score <= 10:
                raise ValueError

            reason = str(dict_content["reason"])

            return score_result.ScoreResult(
                name=name,
                value=score / 10,
                reason=reason,
            )

        else:
            # Compute score using top logprobs
            score_token_position = 3
            log_probs_content = content.choices[0].model_extra["logprobs"]["content"][
                score_token_position
            ]

            top_score_logprobs = log_probs_content["top_logprobs"]
            log_probs_token = log_probs_content["token"]

            linear_probs_sum = 0.0
            weighted_score_sum = 0.0

            for token_info in top_score_logprobs:
                # litellm in v1.60.2 (or earlier) started provide logprobes
                # as pydantic model, not just dict
                # we will convert model to dict to provide backward compatability
                if not isinstance(token_info, dict):
                    token_info = token_info.model_dump()

                # if not a number
                if not token_info["token"].isdecimal():
                    continue

                score = int(token_info["token"])

                # if score value not in scale
                if not 0 <= score <= 10:
                    continue

                log_prob = token_info["logprob"]
                linear_prob = math.exp(log_prob)

                linear_probs_sum += linear_prob
                weighted_score_sum += linear_prob * score

            if linear_probs_sum != 0.0:
                final_score: float = weighted_score_sum / linear_probs_sum / 10
            else:
                # Handle cases where we can't find any matching tokens in the top_log_probs
                if not log_probs_token.isdecimal():
                    raise exceptions.MetricComputationError(GEVAL_SCORE_CALC_FAILED)

                final_score = int(log_probs_token) / 10

            if not (0.0 <= final_score <= 1.0):
                raise ValueError

            # Get the reason
            reason = json.loads(content.choices[0].message.content)["reason"]

            # Return the score and the reason
            return score_result.ScoreResult(name=name, value=final_score, reason=reason)
    except Exception as e:
        LOGGER.error(f"Failed to parse model output: {e}", exc_info=True)
        raise exceptions.MetricComputationError(GEVAL_SCORE_CALC_FAILED)



import abc
from typing import Dict, Any

import pydantic
from opik import dict_utils


class BaseOriginalProviderUsage(pydantic.BaseModel, abc.ABC):
    model_config = pydantic.ConfigDict(extra="allow")
    """
    `extra` values are allowed so that the code continued working if provider adds a new field that
    is not reflected in the SDK yet.
    """

    @abc.abstractmethod
    def to_backend_compatible_flat_dict(self, parent_key_prefix: str) -> Dict[str, int]:
        result = {**self.__dict__}

        result = dict_utils.flatten_dict(
            d=result, delim=".", parent_key=parent_key_prefix
        )

        if self.model_extra is not None:
            model_extra = dict_utils.flatten_dict(
                d=self.model_extra, delim=".", parent_key=parent_key_prefix
            )
            result.update(model_extra)

        result = dict_utils.keep_only_values_of_type(d=result, value_type=int)

        return result

    @classmethod
    @abc.abstractmethod
    def from_original_usage_dict(
        cls, usage: Dict[str, Any]
    ) -> "BaseOriginalProviderUsage":
        pass



import dataclasses
import logging
from typing import Callable, Optional, Dict, Any

from google.adk import models as adk_models
from google.genai import types as genai_types

from ... import llm_usage, LLMProvider
from ...llm_usage import opik_usage


LOGGER = logging.Logger(__name__)


class LlmResponseCreateWrapper:
    def __init__(
        self,
        wrapped_response_create: Callable[
            [genai_types.GenerateContentResponse], adk_models.LlmResponse
        ],
    ) -> None:
        self.wrapped_response_create = wrapped_response_create

    def __call__(
        self, generate_content_response: genai_types.GenerateContentResponse
    ) -> adk_models.LlmResponse:
        return _wrap_llm_response_create(
            generate_content_response, self.wrapped_response_create
        )


@dataclasses.dataclass
class LLMUsageData:
    opik_usage: opik_usage.OpikUsage
    model: Optional[str]
    provider: Optional[str]


def pop_llm_usage_data(**result_dict: Dict[str, Any]) -> Optional[LLMUsageData]:
    """Extracts Opik usage metadata from ADK output and removes it from the result dict."""
    custom_metadata = result_dict.get("custom_metadata", None)
    if custom_metadata is None:
        return None

    opik_usage_metadata = custom_metadata.pop("opik_usage", None)
    if opik_usage_metadata is None:
        return None

    model = custom_metadata.pop("model_version", None)
    provider = custom_metadata.pop("provider", None)
    usage = llm_usage.try_build_opik_usage_or_log_error(
        provider=LLMProvider(provider),
        usage=opik_usage_metadata,
        logger=LOGGER,
        error_message="Failed to log token usage from ADK Gemini call",
    )
    if usage is None:
        return None

    return LLMUsageData(opik_usage=usage, model=model, provider=provider)


def _wrap_llm_response_create(
    generate_content_response: genai_types.GenerateContentResponse,
    wrapped_response_create: Callable[
        [genai_types.GenerateContentResponse], adk_models.LlmResponse
    ],
) -> adk_models.LlmResponse:
    usage_metadata = generate_content_response.usage_metadata
    response = wrapped_response_create(generate_content_response)
    if usage_metadata is None:
        return response

    if response.custom_metadata is None:
        response.custom_metadata = {}

    response.custom_metadata["opik_usage"] = usage_metadata
    response.custom_metadata["provider"] = LLMProvider.GOOGLE_AI
    response.custom_metadata["model_version"] = generate_content_response.model_version

    return response



from typing import Optional, Union, Dict, Any

import pydantic

from opik.types import LLMProvider

from . import opik_usage


class LLMUsageInfo(pydantic.BaseModel):
    provider: Optional[Union[LLMProvider, str]] = None
    model: Optional[str] = None
    usage: Optional[Union[Dict[str, Any], opik_usage.OpikUsage]] = None



from .aggregated_metric import AggregatedMetric
from .heuristics.contains import Contains
from .heuristics.equals import Equals
from .heuristics.is_json import IsJson
from .heuristics.levenshtein_ratio import LevenshteinRatio
from .heuristics.regex_match import RegexMatch
from .heuristics.bleu import SentenceBLEU, CorpusBLEU
from .heuristics.rouge import ROUGE
from .llm_judges.answer_relevance.metric import AnswerRelevance
from .llm_judges.context_precision.metric import ContextPrecision
from .llm_judges.context_recall.metric import ContextRecall
from .llm_judges.g_eval.metric import GEval
from .llm_judges.hallucination.metric import Hallucination
from .llm_judges.moderation.metric import Moderation
from .llm_judges.usefulness.metric import Usefulness
from .base_metric import BaseMetric
from opik.exceptions import MetricComputationError

# from .llm_judges.factuality.metric import Factuality

__all__ = [
    "AggregatedMetric",
    "AnswerRelevance",
    "Contains",
    "ContextPrecision",
    "ContextRecall",
    "Equals",
    # "Factuality",
    "GEval",
    "Hallucination",
    "IsJson",
    "LevenshteinRatio",
    "Moderation",
    "Usefulness",
    "RegexMatch",
    "MetricComputationError",
    "BaseMetric",
    "SentenceBLEU",
    "CorpusBLEU",
    "ROUGE",
]



import os
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
)

import opik.types as opik_types
from opik.api_objects import span
from opik.decorator import arguments_helpers, base_track_decorator

from . import llm_response_wrapper


def convert_adk_base_models(arg: Any) -> Dict[str, Any]:
    """Most ADK objects are Pydantic Base Models"""
    return arg.model_dump(mode="json", exclude_unset=True)


def get_adk_provider() -> opik_types.LLMProvider:
    use_vertexai = os.environ.get("GOOGLE_GENAI_USE_VERTEXAI", "0").lower() in [
        "true",
        "1",
    ]
    return (
        opik_types.LLMProvider.GOOGLE_VERTEXAI
        if use_vertexai
        else opik_types.LLMProvider.GOOGLE_AI
    )


class ADKLLMTrackDecorator(base_track_decorator.BaseTrackDecorator):
    """
    An implementation of BaseTrackDecorator designed specifically for guardrails span.
    """

    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        track_options: arguments_helpers.TrackOptions,
        args: Optional[Tuple],
        kwargs: Optional[Dict[str, Any]],
    ) -> arguments_helpers.StartSpanParameters:
        if args is None:
            raise ValueError("args cannot be None")

        result = arguments_helpers.StartSpanParameters(
            name=track_options.name or "llm",
            input=convert_adk_base_models(args[0]),
            type="llm",
        )

        return result

    def _end_span_inputs_preprocessor(
        self,
        output: Any,
        capture_output: bool,
        current_span_data: span.SpanData,
    ) -> arguments_helpers.EndSpanParameters:
        result_dict = convert_adk_base_models(output)
        usage_data = llm_response_wrapper.pop_llm_usage_data(**result_dict)

        if usage_data is not None:
            result = arguments_helpers.EndSpanParameters(
                output=result_dict,
                usage=usage_data.opik_usage,
                model=usage_data.model,
                provider=get_adk_provider(),
            )
        else:
            result = arguments_helpers.EndSpanParameters(
                output=result_dict,
            )

        return result

    def _streams_handler(
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], str]],
    ) -> Optional[Union[Generator, AsyncGenerator]]:
        return super()._streams_handler(output, capture_output, generations_aggregator)


class ADKToolTrackDecorator(base_track_decorator.BaseTrackDecorator):
    """
    An implementation of BaseTrackDecorator designed specifically for guardrails span.
    """

    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        track_options: arguments_helpers.TrackOptions,
        args: Optional[Tuple],
        kwargs: Optional[Dict[str, Any]],
    ) -> arguments_helpers.StartSpanParameters:
        if args is None:
            raise ValueError("args cannot be None")

        result = arguments_helpers.StartSpanParameters(
            name=track_options.name or "tool",
            input=args[0],
            type="tool",
        )

        return result

    def _end_span_inputs_preprocessor(
        self,
        output: Any,
        capture_output: bool,
        current_span_data: span.SpanData,
    ) -> arguments_helpers.EndSpanParameters:
        result = arguments_helpers.EndSpanParameters(
            output=output,
        )

        return result

    def _streams_handler(
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], str]],
    ) -> Optional[Union[Generator, AsyncGenerator]]:
        return super()._streams_handler(output, capture_output, generations_aggregator)



import uuid
import logging
from typing import Any, Dict, List, Optional

from google.adk.agents.callback_context import CallbackContext
from google.adk.models import LlmRequest, LlmResponse
from google.adk.tools.base_tool import BaseTool
from google.adk.tools.tool_context import ToolContext
from opik import context_storage, datetime_helpers
from opik.api_objects import helpers, opik_client, trace
from opik.decorator import arguments_helpers
from opik.types import DistributedTraceHeadersDict

from . import adk_decorators
from . import llm_response_wrapper

LOGGER = logging.getLogger(__name__)


class OpikTracer:
    def __init__(
        self,
        name: Optional[str] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        project_name: Optional[str] = None,
        distributed_headers: Optional[DistributedTraceHeadersDict] = None,
    ):
        self.name = name
        self.tags = tags
        self.metadata = metadata or {}
        self.project_name = project_name
        self.distributed_headers = distributed_headers
        self.trace_data: Optional[trace.TraceData] = None
        self._client = opik_client.get_client_cached()

        self._last_model_output: Optional[Dict[str, Any]] = None

        self._llm_tracer = adk_decorators.ADKLLMTrackDecorator()
        self._tool_tracer = adk_decorators.ADKToolTrackDecorator()

        # monkey patch LLMResponse to store usage_metadata
        old_function = LlmResponse.create
        create_wrapper = llm_response_wrapper.LlmResponseCreateWrapper(old_function)
        LlmResponse.create = create_wrapper

    def before_agent_callback(
        self, callback_context: CallbackContext, *args: Any, **kwargs: Any
    ) -> None:
        if "opik_thread_id" in callback_context.state:
            thread_id = callback_context.state["opik_thread_id"]
        else:
            thread_id = str(uuid.uuid4())
            callback_context.state["opik_thread_id"] = thread_id

        # Should we create a trace here as we don't have an input?

        trace_metadata = self.metadata.copy()
        trace_metadata["adk_invocation_id"] = callback_context.invocation_id

        user_input = adk_decorators.convert_adk_base_models(
            callback_context.user_content
        )

        self.trace_data = trace.TraceData(
            id=helpers.generate_id(),
            start_time=datetime_helpers.local_timestamp(),
            name=self.name or callback_context.agent_name,
            input=user_input,
            metadata=trace_metadata,
            tags=self.tags,
            project_name=self.project_name,
            thread_id=thread_id,
        )

        context_storage.set_trace_data(self.trace_data)

    def after_agent_callback(
        self, callback_context: CallbackContext, *args: Any, **kwargs: Any
    ) -> None:
        assert context_storage.get_trace_data() is self.trace_data
        assert self.trace_data is not None

        context_storage.pop_trace_data()

        output = self._last_model_output

        # remove the custom metadata with opik usage we added
        if output is not None:
            llm_response_wrapper.pop_llm_usage_data(**output)

        self.trace_data.update(output=output).init_end_time()

        self._client.trace(**self.trace_data.__dict__)

        # Cleaning
        self.trace_data = None
        self._last_model_output = None

    def before_model_callback(
        self,
        callback_context: CallbackContext,
        llm_request: LlmRequest,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._llm_tracer._before_call(
            func=self.before_model_callback,
            track_options=arguments_helpers.TrackOptions(
                name=llm_request.model,
                type="llm",
                capture_input=True,
                capture_output=True,
                flush=False,
                tags=[],
                metadata={},
                ignore_arguments=None,
                generations_aggregator=None,
                project_name=self.project_name,
            ),
            args=(llm_request,),
            kwargs=kwargs,
        )

    def after_model_callback(
        self,
        callback_context: CallbackContext,
        llm_response: LlmResponse,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        try:
            # Ignore partial chunks, ADK will call this method with the full
            # response at the end
            if llm_response.partial is True:
                return
        except Exception:
            LOGGER.debug("Error checking for partial chunks", exc_info=True)

        self._last_model_output = adk_decorators.convert_adk_base_models(llm_response)

        self._llm_tracer._after_call(
            output=llm_response,
            error_info=None,
            capture_output=True,
            generators_span_to_end=None,
            generators_trace_to_end=None,
            flush=False,
        )

    def before_tool_callback(
        self,
        tool: BaseTool,
        args: Dict[str, Any],
        tool_context: ToolContext,
        *other_args: Any,
        **kwargs: Any,
    ) -> None:
        metadata = {"function_call_id": tool_context.function_call_id}

        self._tool_tracer._before_call(
            func=self.before_model_callback,
            track_options=arguments_helpers.TrackOptions(
                name=tool.name,
                type="tool",
                capture_input=True,
                capture_output=True,
                flush=False,
                tags=[],
                metadata=metadata,
                ignore_arguments=None,
                generations_aggregator=None,
                project_name=self.project_name,
            ),
            args=(args,),
            kwargs=kwargs,
        )

    def after_tool_callback(
        self,
        tool: BaseTool,
        args: Dict[str, Any],
        tool_context: ToolContext,
        tool_response: Dict,
        *other_args: Any,
        **kwargs: Any,
    ) -> None:
        self._tool_tracer._after_call(
            output=tool_response,
            error_info=None,
            capture_output=True,
            generators_span_to_end=None,
            generators_trace_to_end=None,
            flush=False,
        )



import pydantic
from typing import Optional, Dict, Any
from . import base_original_provider_usage
from opik import dict_utils


class OpenAICompletionsUsage(base_original_provider_usage.BaseOriginalProviderUsage):
    """OpenAI calls token usage data (or token usage data in OpenAI format). Updated 11.03.2025"""

    completion_tokens: int
    """Number of tokens in the generated completion."""

    prompt_tokens: int
    """Number of tokens in the prompt."""

    total_tokens: int
    """Total number of tokens used in the request (prompt + completion)."""

    completion_tokens_details: Optional["CompletionTokensDetails"] = None
    """Breakdown of tokens used in a completion."""

    prompt_tokens_details: Optional["PromptTokensDetails"] = None
    """Breakdown of tokens used in the prompt."""

    def to_backend_compatible_flat_dict(self, parent_key_prefix: str) -> Dict[str, int]:
        result = {**self.__dict__}

        if self.completion_tokens_details is not None:
            result["completion_tokens_details"] = (
                self.completion_tokens_details.model_dump()
            )

        if self.prompt_tokens_details is not None:
            result["prompt_tokens_details"] = self.prompt_tokens_details.model_dump()

        result = dict_utils.flatten_dict(
            d=result, delim=".", parent_key=parent_key_prefix
        )

        if self.model_extra is not None:
            model_extra = dict_utils.flatten_dict(
                d=self.model_extra, delim=".", parent_key=parent_key_prefix
            )
            result.update(model_extra)

        result = dict_utils.keep_only_values_of_type(d=result, value_type=int)
        return result

    @classmethod
    def from_original_usage_dict(
        cls, usage_dict: Dict[str, Any]
    ) -> "OpenAICompletionsUsage":
        usage_dict = {**usage_dict}
        completion_tokens_details_raw = usage_dict.pop(
            "completion_tokens_details", None
        )
        prompt_tokens_details_raw = usage_dict.pop("prompt_tokens_details", None)

        completion_tokens_details = (
            CompletionTokensDetails(**completion_tokens_details_raw)
            if isinstance(completion_tokens_details_raw, dict)
            else None
        )

        prompt_tokens_details = (
            PromptTokensDetails(**prompt_tokens_details_raw)
            if isinstance(prompt_tokens_details_raw, dict)
            else None
        )

        return cls(
            **usage_dict,
            completion_tokens_details=completion_tokens_details,
            prompt_tokens_details=prompt_tokens_details,
        )


class CompletionTokensDetails(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")

    accepted_prediction_tokens: Optional[int] = None
    """
    When using Predicted Outputs, the number of tokens in the prediction that
    appeared in the completion.
    """

    audio_tokens: Optional[int] = None
    """Audio input tokens generated by the model."""

    reasoning_tokens: Optional[int] = None
    """Tokens generated by the model for reasoning."""

    rejected_prediction_tokens: Optional[int] = None
    """
    When using Predicted Outputs, the number of tokens in the prediction that did
    not appear in the completion. However, like reasoning tokens, these tokens are
    still counted in the total completion tokens for purposes of billing, output,
    and context window limits.
    """


class PromptTokensDetails(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")

    audio_tokens: Optional[int] = None
    """Audio input tokens present in the prompt."""

    cached_tokens: Optional[int] = None
    """Cached tokens present in the prompt."""



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .types import OptimizationUpdateStatus, OptimizationWriteStatus

__all__ = ["OptimizationUpdateStatus", "OptimizationWriteStatus"]



import base64
import dataclasses
import datetime as dt
import logging
from enum import Enum
from pathlib import PurePath
from types import GeneratorType
from typing import Any, Callable, Optional, Set, Tuple, Type

import pydantic

import opik.rest_api.core.datetime_utils as datetime_utils

try:
    import numpy as np
except ImportError:
    np = None

LOGGER = logging.getLogger(__name__)

_ENCODER_EXTENSIONS: Set[Tuple[Type, Callable[[Any], Any]]] = set()


def register_encoder_extension(obj_type: Type, encoder: Callable[[Any], Any]) -> None:
    _ENCODER_EXTENSIONS.add((obj_type, encoder))


def encode(obj: Any, seen: Optional[Set[int]] = None) -> Any:
    """
    This is a modified version of the serializer generated by Fern in rest_api.core.jsonable_encoder.
    The code is simplified to serialize complex objects into a textual representation.
    It also handles cyclic references to avoid infinite recursion.
    """
    if seen is None:
        seen = set()

    if hasattr(obj, "__dict__"):
        obj_id = id(obj)
        if obj_id in seen:
            LOGGER.debug(f"Found cyclic reference to {type(obj).__name__} id={obj_id}")
            return f"<Cyclic reference to {type(obj).__name__} id={obj_id}>"
        seen.add(obj_id)

    try:
        for type_, encoder in _ENCODER_EXTENSIONS:
            if isinstance(obj, type_):
                return encode(encoder(obj), seen)

        if dataclasses.is_dataclass(obj) or isinstance(obj, pydantic.BaseModel):
            obj_dict = obj.__dict__
            return encode(obj_dict, seen)

        if isinstance(obj, Enum):
            return encode(obj.value, seen)
        if isinstance(obj, PurePath):
            return str(obj)
        if isinstance(obj, (str, int, float, type(None))):
            return obj
        if isinstance(obj, dt.datetime):
            return datetime_utils.serialize_datetime(obj)
        if isinstance(obj, dt.date):
            return str(obj)
        if isinstance(obj, bytes):
            return base64.b64encode(obj).decode("utf-8")
        if isinstance(obj, dict):
            encoded_dict = {}
            allowed_keys = set(obj.keys())
            for key, value in obj.items():
                if key in allowed_keys:
                    encoded_key = encode(key, seen)
                    encoded_value = encode(value, seen)
                    encoded_dict[encoded_key] = encoded_value
            return encoded_dict
        if isinstance(obj, (list, set, frozenset, GeneratorType, tuple)):
            encoded_list = []
            for item in obj:
                encoded_list.append(encode(item, seen))
            return encoded_list

        if np is not None and isinstance(obj, np.ndarray):
            return encode(obj.tolist(), seen)

    except Exception:
        LOGGER.debug("Failed to serialize object.", exc_info=True)

    finally:
        # Once done encoding this object, remove from `seen`,
        # so the same object can appear again at a sibling branch.
        if hasattr(obj, "__dict__"):
            obj_id = id(obj)
            seen.remove(obj_id)

    data = str(obj)

    return data



from typing import Any, Dict, Optional

import haystack
from haystack import logging, tracing

from . import opik_tracer
import opik

LOGGER = logging.getLogger(__name__)


@haystack.component
class OpikConnector:
    """
    OpikConnector connects Haystack LLM framework with [Opik](https://github.com/comet-ml/opik) in order to enable the
    tracing of operations and data flow within various components of a pipeline.

    Simply add this component to your pipeline, but *do not* connect it to any other component. The OpikConnector
    will automatically trace the operations and data flow within the pipeline.

    In order to configure Opik, you will need to call first install the Opik SDK using `pip install opik` and then
    run `opik configure` from the command line. Alternatively you can configure Opik using environment variables,
    you can find more information about how to configure Opik [here](https://www.comet.com/docs/opik/tracing/sdk_configuration).

    In addition, you need to set the `HAYSTACK_CONTENT_TRACING_ENABLED` environment variable to `true` in order to
    enable Haystack tracing in your pipeline.

    Example:
        You can use the `OpikConnector` in the following way::

                import os

                os.environ["HAYSTACK_CONTENT_TRACING_ENABLED"] = "true"

                from haystack import Pipeline
                from haystack.components.builders import ChatPromptBuilder
                from haystack.components.generators.chat import OpenAIChatGenerator
                from haystack.dataclasses import ChatMessage
                from opik.integrations.haystack import (
                    OpikConnector,
                )

                if __name__ == "__main__":
                    pipe = Pipeline()
                    pipe.add_component("tracer", OpikConnector("Chat example"))
                    pipe.add_component("prompt_builder", ChatPromptBuilder())
                    pipe.add_component("llm", OpenAIChatGenerator(model="gpt-3.5-turbo"))

                    pipe.connect("prompt_builder.prompt", "llm.messages")

                    messages = [
                        ChatMessage.from_system(
                            "Always respond in German even if some input data is in other languages."
                        ),
                        ChatMessage.from_user("Tell me about {{location}}"),
                    ]

                    response = pipe.run(
                        data={
                            "prompt_builder": {
                                "template_variables": {"location": "Berlin"},
                                "template": messages,
                            }
                        }
                    )
                    print(response["llm"]["replies"][0])
                    print(response["tracer"]["trace_url"])

    Note:

        You may disable flushing the data after each component by setting the `HAYSTACK_OPIK_ENFORCE_FLUSH`
        environent variable to `false`. By default, the data is flushed after each component and blocks the thread until
        the data is sent to Opik. **Caution**: Disabling this feature may result in data loss if the program crashes
        before the data is sent to Opik. Make sure you will call the `flush()` method explicitly before the program exits.
        E.g. by using tracer.actual_tracer.flush()::

            from haystack.tracing import tracer

            tracer.actual_tracer.flush()
    """

    def __init__(self, name: str, project_name: Optional[str] = None):
        """
        Initialize the OpikConnector component.

        Args:
            name: The name of the pipeline or component. This name will be used to identify the tracing run on the
                Opik dashboard.
            project_name: The name of the project to use for the tracing run. If not provided, the project name will be
                set to the default project name.
        """
        self.name = name
        self.tracer = opik_tracer.OpikTracer(
            opik_client=opik.Opik(project_name=project_name), name=name
        )
        tracing.enable_tracing(self.tracer)

    @haystack.component.output_types(name=str, trace_id=Optional[str], project_url=str)
    def run(
        self, invocation_context: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Runs the OpikConnector component.

        Args:
            invocation_context: A dictionary with additional context for the invocation. This parameter
                is useful when users want to mark this particular invocation with additional information, e.g.
                a run id from their own execution framework, user id, etc. These key-value pairs are then visible
                in the Opik traces.

        Returns:
            A dictionary with the following keys:
                - `name`: The name of the tracing component.
                - `trace_id`: The Opik trace id.
                - `project_url`: The URL to the Opik project with tracing data.
        """
        LOGGER.debug(
            f"Opik tracer invoked with the following context: {invocation_context}"
        )

        return {
            "name": self.name,
            "trace_id": self.tracer.get_trace_id(),
            "project_url": self.tracer.get_project_url(),
        }



import logging
from typing import TYPE_CHECKING, Any, Dict, Literal, Optional, Tuple, Union

from opik import llm_usage, logging_messages
from opik.types import LLMProvider

if TYPE_CHECKING:
    from langchain_core.tracers.schemas import Run

LOGGER = logging.getLogger(__name__)


def get_llm_usage_info(
    run_dict: Optional[Dict[str, Any]] = None,
) -> llm_usage.LLMUsageInfo:
    if run_dict is None:
        return llm_usage.LLMUsageInfo()

    usage_dict = _try_get_token_usage(run_dict)
    provider, model = _get_provider_and_model(run_dict)

    return llm_usage.LLMUsageInfo(provider=provider, model=model, usage=usage_dict)


def _try_get_token_usage(run_dict: Dict[str, Any]) -> Optional[llm_usage.OpikUsage]:
    try:
        # TODO: This is empty in Streaming mode with gemini models
        usage_metadata = run_dict["outputs"]["generations"][-1][-1]["generation_info"][
            "usage_metadata"
        ]

        opik_usage = llm_usage.OpikUsage.from_google_dict(usage_metadata)
        return opik_usage
    except Exception:
        LOGGER.warning(
            logging_messages.FAILED_TO_EXTRACT_TOKEN_USAGE_FROM_PRESUMABLY_LANGCHAIN_GOOGLE_LLM_RUN,
            exc_info=True,
        )
        return None


def is_google_run(run: "Run") -> bool:
    try:
        if run.serialized is None:
            return False

        invocation_params = run.extra.get("invocation_params", {})
        provider = invocation_params.get("_type", "").lower()
        is_google = "vertexai" in provider.lower()

        return is_google

    except Exception:
        LOGGER.debug(
            "Failed to check if Run instance is from Google LLM, returning False.",
            exc_info=True,
        )
        return False


def _get_provider_and_model(
    run_dict: Dict[str, Any],
) -> Tuple[Optional[Union[Literal[LLMProvider.GOOGLE_VERTEXAI], str]], Optional[str]]:
    """
    Fetches the provider and model information from a given run dictionary.
    """
    provider = None
    model = None

    if invocation_params := run_dict["extra"].get("invocation_params"):
        provider = invocation_params.get("_type")
        if provider == "vertexai":
            provider = LLMProvider.GOOGLE_VERTEXAI
        model = invocation_params.get("model_name")

    return provider, model



import logging
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

import openai
from openai import _types as _openai_types
from openai.types.chat import chat_completion, chat_completion_chunk
from typing_extensions import override

from opik import dict_utils, llm_usage
from opik.api_objects import span
from opik.decorator import arguments_helpers, base_track_decorator
from opik.integrations.openai import chat_completion_chunks_aggregator
from opik.types import LLMProvider

from . import stream_patchers

LOGGER = logging.getLogger(__name__)

KWARGS_KEYS_TO_LOG_AS_INPUTS = ["messages", "function_call"]
RESPONSE_KEYS_TO_LOG_AS_OUTPUT = ["choices"]


class OpenaiChatCompletionsTrackDecorator(base_track_decorator.BaseTrackDecorator):
    """
    An implementation of BaseTrackDecorator designed specifically for tracking
    calls of OpenAI's `chat.completion.create` and `chat.completions.parse` functions.

    Besides special processing for input arguments and response content, it
    overrides _generators_handler() method to work correctly with
    openai.Stream and openai.AsyncStream objects.
    """

    def __init__(self) -> None:
        super().__init__()
        self.provider = "openai"

    @override
    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        track_options: arguments_helpers.TrackOptions,
        args: Optional[Tuple],
        kwargs: Optional[Dict[str, Any]],
    ) -> arguments_helpers.StartSpanParameters:
        assert (
            kwargs is not None
        ), "Expected kwargs to be not None in chat.completion.create(**kwargs), chat.completion.parse(**kwargs) or chat.completion.stream(**kwargs)"

        name = track_options.name if track_options.name is not None else func.__name__
        if _is_completions_stream_call(
            name_passed_to_track_decorator=name, kwargs=kwargs
        ):
            kwargs = _remove_not_given_sentinel_values(kwargs)
            name = "chat_completion_stream"

        metadata = track_options.metadata if track_options.metadata is not None else {}

        input, new_metadata = dict_utils.split_dict_by_keys(
            kwargs, keys=KWARGS_KEYS_TO_LOG_AS_INPUTS
        )
        metadata = dict_utils.deepmerge(metadata, new_metadata)
        metadata.update(
            {
                "created_from": "openai",
                "type": "openai_chat",
            }
        )

        tags = ["openai"]

        result = arguments_helpers.StartSpanParameters(
            name=name,
            input=input,
            type=track_options.type,
            tags=tags,
            metadata=metadata,
            project_name=track_options.project_name,
            model=kwargs.get("model", None),
            provider=self.provider,
        )

        return result

    @override
    def _end_span_inputs_preprocessor(
        self,
        output: Any,
        capture_output: bool,
        current_span_data: span.SpanData,
    ) -> arguments_helpers.EndSpanParameters:
        assert isinstance(
            output,
            (
                chat_completion.ChatCompletion,
                chat_completion_chunks_aggregator.ChatCompletionChunksAggregated,
            ),
        )

        result_dict = output.model_dump(mode="json")
        output, metadata = dict_utils.split_dict_by_keys(result_dict, ["choices"])

        opik_usage = None
        if "usage" in result_dict:
            opik_usage = llm_usage.try_build_opik_usage_or_log_error(
                provider=LLMProvider.OPENAI,
                usage=result_dict["usage"],
                logger=LOGGER,
                error_message="Failed to log token usage from openai call",
            )

        model = result_dict["model"]

        result = arguments_helpers.EndSpanParameters(
            output=output,
            usage=opik_usage,
            metadata=metadata,
            model=model,
            provider=self.provider,
        )

        return result

    @override
    def _streams_handler(  # type: ignore
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[
            Callable[
                [List[chat_completion_chunk.ChatCompletionChunk]],
                chat_completion_chunks_aggregator.ChatCompletionChunksAggregated,
            ]
        ],
    ) -> Union[
        None,
        Iterator[chat_completion_chunk.ChatCompletionChunk],
        AsyncIterator[chat_completion_chunk.ChatCompletionChunk],
    ]:
        assert (
            generations_aggregator is not None
        ), "OpenAI decorator will always get aggregator function as input"

        if isinstance(output, openai.Stream):
            span_to_end, trace_to_end = base_track_decorator.pop_end_candidates()
            return stream_patchers.patch_sync_stream(
                stream=output,
                span_to_end=span_to_end,
                trace_to_end=trace_to_end,
                generations_aggregator=generations_aggregator,
                finally_callback=self._after_call,
            )

        if isinstance(output, openai.AsyncStream):
            span_to_end, trace_to_end = base_track_decorator.pop_end_candidates()
            return stream_patchers.patch_async_stream(
                stream=output,
                span_to_end=span_to_end,
                trace_to_end=trace_to_end,
                generations_aggregator=generations_aggregator,
                finally_callback=self._after_call,
            )

        NOT_A_STREAM = None

        return NOT_A_STREAM


def _remove_not_given_sentinel_values(dict_: Dict[str, Any]) -> Dict[str, Any]:
    """
    Under the hood of `stream()` method openai calls `create(..,stream=True,..)`
    and passes a lot of NOT_GIVEN values that we don't want to track.
    """
    return {
        key: value
        for key, value in dict_.items()
        if value is not _openai_types.NOT_GIVEN
    }


def _is_completions_stream_call(
    name_passed_to_track_decorator: str, kwargs: Dict[str, Any]
) -> bool:
    if not name_passed_to_track_decorator == "chat_completion_create":
        return False

    for _, value in kwargs.items():
        if value is _openai_types.NOT_GIVEN:
            return True

    return False



import contextvars
import contextlib

from typing import List, Optional, Generator
from opik.api_objects import span, trace


class OpikContextStorage:
    """
    Class responsible for keeping and providing access to the context of
    spans and traces.

    Read if you are going to change this class.

    If you want to add or pop the span from the stack
    you must NOT do it like that: spans_stack_context.get().append()

    If you modify the default object in ContextVar (which is a MUTABLE list)
    these changes will be reflected in any new context which breaks the whole idea
    of ContextVar.

    You should append to it like that:
    span_data_stack = span_data_stack_context.get().copy()
    span_data_stack_context.set(span_data_stack + [new_element])

    And pop like that:
    span_data_stack = spans_stack_context.get().copy()
    spans_stack_context.set(span_data_stack[:-1])

    The following functions provide an API to work with ContextVars this way
    """

    def __init__(self) -> None:
        self._current_trace_data_context: contextvars.ContextVar[
            Optional[trace.TraceData]
        ] = contextvars.ContextVar("current_trace_data", default=None)
        self._spans_data_stack_context: contextvars.ContextVar[List[span.SpanData]] = (
            contextvars.ContextVar("spans_data_stack", default=[])
        )

    def top_span_data(self) -> Optional[span.SpanData]:
        if self.span_data_stack_empty():
            return None
        stack = self._get_data_stack()
        return stack[-1]

    def pop_span_data(self) -> Optional[span.SpanData]:
        stack = self._get_data_stack()
        self._spans_data_stack_context.set(stack[:-1])
        return stack[-1]

    def add_span_data(self, span: span.SpanData) -> None:
        stack = self._get_data_stack()
        self._spans_data_stack_context.set(stack + [span])

    def span_data_stack_empty(self) -> bool:
        return len(self._get_data_stack()) == 0

    def _get_data_stack(self) -> List[span.SpanData]:
        return self._spans_data_stack_context.get().copy()

    def get_trace_data(self) -> Optional[trace.TraceData]:
        trace = self._current_trace_data_context.get()
        return trace

    def pop_trace_data(self) -> Optional[trace.TraceData]:
        trace = self._current_trace_data_context.get()
        self.set_trace_data(None)
        return trace

    def set_trace_data(self, trace: Optional[trace.TraceData]) -> None:
        self._current_trace_data_context.set(trace)

    def clear_all(self) -> None:
        self._current_trace_data_context.set(None)
        self._spans_data_stack_context.set([])


_context_storage = OpikContextStorage()

top_span_data = _context_storage.top_span_data
pop_span_data = _context_storage.pop_span_data
add_span_data = _context_storage.add_span_data
span_data_stack_empty = _context_storage.span_data_stack_empty
get_trace_data = _context_storage.get_trace_data
pop_trace_data = _context_storage.pop_trace_data
set_trace_data = _context_storage.set_trace_data
clear_all = _context_storage.clear_all


@contextlib.contextmanager
def temporary_context(
    span_data: span.SpanData, trace_data: Optional[trace.TraceData]
) -> Generator[None, None, None]:
    """
    Temporary adds span and trace data to the context.
    If trace_data is None, it has no effect on the current trace in the context.
    """
    try:
        original_trace = get_trace_data()

        if trace_data is not None:
            set_trace_data(trace=trace_data)

        add_span_data(span_data)

        yield
    finally:
        set_trace_data(original_trace)
        pop_span_data()



from typing import Any, Dict, List, Optional, Union

from opik import llm_usage
from opik.api_objects import span, trace
from opik.api_objects.attachment import Attachment
from opik.types import DistributedTraceHeadersDict, FeedbackScoreDict, LLMProvider

from . import context_storage, exceptions


def get_current_span_data() -> Optional[span.SpanData]:
    """
    Returns current span created by track() decorator or None if no span was found.
    """
    span_data = context_storage.top_span_data()
    if span_data is None:
        return None

    return span.SpanData(**span_data.__dict__)


def get_current_trace_data() -> Optional[trace.TraceData]:
    """
    Returns current trace created by track() decorator or None if no trace was found.
    """
    trace_data = context_storage.get_trace_data()
    if trace_data is None:
        return None

    return trace.TraceData(**trace_data.__dict__)


def get_distributed_trace_headers() -> DistributedTraceHeadersDict:
    """
    Returns headers dictionary to be passed into tracked function on remote node.
    Requires an existing span in the context, otherwise raises an error.
    """
    current_span_data = context_storage.top_span_data()

    if current_span_data is None:
        raise Exception("There is no span in the context.")

    return DistributedTraceHeadersDict(
        opik_trace_id=current_span_data.trace_id,
        opik_parent_span_id=current_span_data.id,
    )


def update_current_span(
    name: Optional[str] = None,
    input: Optional[Dict[str, Any]] = None,
    output: Optional[Dict[str, Any]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
    usage: Optional[Union[Dict[str, Any], llm_usage.OpikUsage]] = None,
    feedback_scores: Optional[List[FeedbackScoreDict]] = None,
    model: Optional[str] = None,
    provider: Optional[Union[str, LLMProvider]] = None,
    total_cost: Optional[float] = None,
    attachments: Optional[List[Attachment]] = None,
) -> None:
    """
    Update the current span with the provided parameters. This method is usually called within a tracked function.

    Args:
        name: The name of the span.
        input: The input data of the span.
        output: The output data of the span.
        metadata: The metadata of the span.
        tags: The tags of the span.
        usage: Usage data for the span. In order for input, output and total tokens to be visible in the UI,
            the usage must contain OpenAI-formatted keys (they can be passed additionally to the original usage on the top level of the dict): prompt_tokens, completion_tokens and total_tokens.
            If OpenAI-formatted keys were not found, Opik will try to calculate them automatically if the usage
            format is recognized (you can see which provider's formats are recognized in opik.LLMProvider enum), but it is not guaranteed.
        feedback_scores: The feedback scores of the span.
        model: The name of LLM (in this case type parameter should be == llm)
        provider: The provider of LLM. You can find providers officially supported by Opik for cost tracking
            in `opik.LLMProvider` enum. If your provider is not here, please open an issue in our GitHub - https://github.com/comet-ml/opik.
            If your provider is not in the list, you can still specify it, but the cost tracking will not be available
        total_cost: The cost of the span in USD. This value takes priority over the cost calculated by Opik from the usage.
        attachments: The list of attachments to be uploaded to the span.
    """
    new_params = {
        "name": name,
        "input": input,
        "output": output,
        "metadata": metadata,
        "tags": tags,
        "usage": usage,
        "feedback_scores": feedback_scores,
        "model": model,
        "provider": provider,
        "total_cost": total_cost,
        "attachments": attachments,
    }
    current_span_data = context_storage.top_span_data()
    if current_span_data is None:
        raise exceptions.OpikException("There is no span in the context.")

    current_span_data.update(**new_params)


def update_current_trace(
    name: Optional[str] = None,
    input: Optional[Dict[str, Any]] = None,
    output: Optional[Dict[str, Any]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
    feedback_scores: Optional[List[FeedbackScoreDict]] = None,
    thread_id: Optional[str] = None,
    attachments: Optional[List[Attachment]] = None,
) -> None:
    """
    Update the current trace with the provided parameters. This method is usually called within a tracked function.

    Args:
        name: The name of the trace.
        input: The input data of the trace.
        output: The output data of the trace.
        metadata: The metadata of the trace.
        tags: The tags of the trace.
        feedback_scores: The feedback scores of the trace.
        thread_id: Used to group multiple traces into a thread.
            The identifier is user-defined and has to be unique per project.
        attachments: The list of attachments to be uploaded to the trace.
    """
    new_params = {
        "name": name,
        "input": input,
        "output": output,
        "metadata": metadata,
        "tags": tags,
        "feedback_scores": feedback_scores,
        "thread_id": thread_id,
        "attachments": attachments,
    }
    current_trace_data = context_storage.get_trace_data()
    if current_trace_data is None:
        raise exceptions.OpikException("There is no trace in the context.")

    current_trace_data.update(**new_params)


__all__ = [
    "get_current_span_data",
    "get_current_trace_data",
    "update_current_span",
    "update_current_trace",
    "get_distributed_trace_headers",
]



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..errors.forbidden_error import ForbiddenError
from ..errors.unauthorized_error import UnauthorizedError
from ..types.auth_details_holder import AuthDetailsHolder
from ..types.error_message import ErrorMessage
from ..types.workspace_name_holder import WorkspaceNameHolder

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawCheckClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def access(
        self, *, request: AuthDetailsHolder, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Check user access to workspace

        Parameters
        ----------
        request : AuthDetailsHolder

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/auth",
            method="POST",
            json=request,
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_workspace_name(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[WorkspaceNameHolder]:
        """
        User's default workspace name

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[WorkspaceNameHolder]
            Authentication resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/auth/workspace",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    WorkspaceNameHolder,
                    parse_obj_as(
                        type_=WorkspaceNameHolder,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawCheckClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def access(
        self, *, request: AuthDetailsHolder, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Check user access to workspace

        Parameters
        ----------
        request : AuthDetailsHolder

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/auth",
            method="POST",
            json=request,
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_workspace_name(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[WorkspaceNameHolder]:
        """
        User's default workspace name

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[WorkspaceNameHolder]
            Authentication resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/auth/workspace",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    WorkspaceNameHolder,
                    parse_obj_as(
                        type_=WorkspaceNameHolder,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 401:
                raise UnauthorizedError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 403:
                raise ForbiddenError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



from typing import Optional, Any

from .litellm import litellm_chat_model
from . import base_model

DEFAULT_GPT_MODEL_NAME = "gpt-4o"


def get(model_name: Optional[str], **model_kwargs: Any) -> base_model.OpikBaseModel:
    if model_name is None:
        model_name = DEFAULT_GPT_MODEL_NAME

    return litellm_chat_model.LiteLLMChatModel(model_name=model_name, **model_kwargs)



import dataclasses
import logging
import pathlib
from typing import List, Optional, IO

import httpx

from . import file_parts_strategy, s3_upload_error, s3_httpx_client
from .. import file_upload_monitor


LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class PartMetadata:
    e_tag: str
    part_number: int
    size: int


class S3FileDataUploader:
    """Represents S3 file data uploader to be used for sending file data to AWS S3 bucket."""

    def __init__(
        self,
        file_parts: file_parts_strategy.FilePartsStrategy,
        pre_sign_urls: List[str],
        httpx_client: httpx.Client,
        monitor: Optional[file_upload_monitor.FileUploadMonitor] = None,
    ) -> None:
        self._file_parts = file_parts
        self._pre_sign_urls = pre_sign_urls
        self._httpx_client = httpx_client
        self._monitor = monitor

        self.uploaded_parts: List[PartMetadata] = []
        self.bytes_sent = 0

    def upload(self) -> List[PartMetadata]:
        self._file_parts.calculate()
        file_to_upload = pathlib.Path(self._file_parts.file)
        try:
            with file_to_upload.open("rb") as fp:
                self._upload(fp=fp)
        except Exception as e:
            raise s3_upload_error.S3UploadFileError(
                file=self._file_parts.file, reason=str(e)
            ) from e

        return self.uploaded_parts

    def _upload(self, fp: IO) -> None:
        max_file_part_size = self._file_parts.max_file_part_size
        for i, url in enumerate(self._pre_sign_urls):
            part_number = i + 1
            file_data = fp.read(max_file_part_size)
            self._send_data_part(url=url, file_data=file_data, part_number=part_number)

    @s3_httpx_client.s3_retry
    def _send_data_part(self, url: str, file_data: bytes, part_number: int) -> None:
        response = self._httpx_client.put(
            url=url,
            content=file_data,
        )
        if response.status_code == httpx.codes.OK:
            e_tag = response.headers["ETag"]
            self._on_part_complete(
                PartMetadata(e_tag=e_tag, part_number=part_number, size=len(file_data))
            )

        response.raise_for_status()

        LOGGER.debug(
            "Part [%04d / %04d] of file '%s' was uploaded to S3",
            part_number,
            len(self._pre_sign_urls),
            self._file_parts.file,
        )

    def _on_part_complete(self, part: PartMetadata) -> None:
        self.uploaded_parts.append(part)
        self.bytes_sent += part.size
        if self._monitor is not None:
            self._monitor.update(part.size)



import threading
import time
import abc

from typing import List, Callable
from .. import messages


class BaseBatcher(abc.ABC):
    def __init__(
        self,
        flush_callback: Callable[[messages.BaseMessage], None],
        max_batch_size: int,
        flush_interval_seconds: float,
    ):
        self._flush_interval_seconds: float = flush_interval_seconds
        self._flush_callback: Callable[[messages.BaseMessage], None] = flush_callback
        self._accumulated_messages: List[messages.BaseMessage] = []
        self._max_batch_size: int = max_batch_size

        self._last_time_flush_callback_called: float = time.time()
        self._lock = threading.RLock()

    def flush(self) -> None:
        with self._lock:
            if len(self._accumulated_messages) > 0:
                batch_message = self._create_batch_from_accumulated_messages()
                self._accumulated_messages = []

                self._flush_callback(batch_message)
            self._last_time_flush_callback_called = time.time()

    def is_ready_to_flush(self) -> bool:
        return (
            time.time() - self._last_time_flush_callback_called
        ) >= self._flush_interval_seconds

    def is_empty(self) -> bool:
        with self._lock:
            return len(self._accumulated_messages) == 0

    @abc.abstractmethod
    def _create_batch_from_accumulated_messages(self) -> messages.BaseMessage: ...

    @abc.abstractmethod
    def add(self, message: messages.BaseMessage) -> None:
        with self._lock:
            self._accumulated_messages.append(message)
            if len(self._accumulated_messages) >= self._max_batch_size:
                self.flush()



import logging
import os
import httpx

from typing import Any

import opik.hooks

LOGGER = logging.getLogger(__name__)


def _in_aws_sagemaker() -> bool:
    return os.getenv("AWS_PARTNER_APP_AUTH") is not None


class SagemakerAuth(httpx.Auth):
    def __init__(self, auth_provider: Any) -> None:
        self.auth_provider = auth_provider

    def auth_flow(self, request):  # type: ignore
        if not _in_aws_sagemaker():
            yield request

        url, signed_headers = self.auth_provider.get_signed_request(
            str(request.url), request.method, request.headers, request.content
        )

        request.url = httpx.URL(url)
        request.headers.update(signed_headers)

        yield request


def setup_aws_sagemaker_session_hook() -> None:
    def sagemaker_auth_client_hook(client: httpx.Client) -> None:
        if not _in_aws_sagemaker():
            return

        import sagemaker

        auth_provider = sagemaker.PartnerAppAuthProvider()
        sagemaker_auth = SagemakerAuth(auth_provider)

        client.auth = sagemaker_auth

    opik.hooks.register_httpx_client_hook(sagemaker_auth_client_hook)
